import torch
from torch import nn

import cheetah

from .resources import ARESlatticeStage3v1_9 as ares


def test_simple_quadrupole():
    """
    Simple test on a [D, Q, D] lattice with the qudrupole's k1 requiring grad, checking
    if PyTorch tracked a grad_fn into the outgoing beam.
    """
    segment = cheetah.Segment(
        [
            cheetah.Drift(length=torch.tensor(1.0)),
            cheetah.Quadrupole(
                length=torch.tensor(0.2),
                k1=nn.Parameter(torch.tensor(3.142)),
                name="my_quad",
            ),
            cheetah.Drift(length=torch.tensor(1.0)),
        ]
    )
    incoming_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/astra/ACHIP_EA1_2021.1351.001"
    )

    outgoing_beam = segment.track(incoming_beam)

    assert outgoing_beam.particles.grad_fn is not None


def test_ea_magnets():
    """
    Test that gradients are tracking when the magnet settings in the ARES experimental
    area require grad.
    """
    ea = cheetah.Segment.from_ocelot(ares.cell, warnings=False).subcell(
        "AREASOLA1", "AREABSCR1"
    )
    incoming_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/astra/ACHIP_EA1_2021.1351.001"
    )

    ea.AREAMQZM2.k1 = nn.Parameter(ea.AREAMQZM2.k1)
    ea.AREAMQZM1.k1 = nn.Parameter(ea.AREAMQZM1.k1)
    ea.AREAMCVM1.angle = nn.Parameter(ea.AREAMCVM1.angle)
    ea.AREAMQZM3.k1 = nn.Parameter(ea.AREAMQZM3.k1)
    ea.AREAMCHM1.angle = nn.Parameter(ea.AREAMCHM1.angle)

    outgoing_beam = ea.track(incoming_beam)

    assert outgoing_beam.particles.grad_fn is not None


def test_ea_incoming_parameter_beam():
    """
    Test that gradients are tracking when incoming beam (being a `ParameterBeam`)
    requires grad.
    """
    ea = cheetah.Segment.from_ocelot(ares.cell, warnings=False).subcell(
        "AREASOLA1", "AREABSCR1"
    )
    incoming_beam = cheetah.ParameterBeam.from_astra(
        "benchmark/astra/ACHIP_EA1_2021.1351.001"
    )

    incoming_beam._mu = nn.Parameter(incoming_beam._mu)
    incoming_beam._cov = nn.Parameter(incoming_beam._cov)

    outgoing_beam = ea.track(incoming_beam)

    assert outgoing_beam._mu.grad_fn is not None
    assert outgoing_beam._cov.grad_fn is not None


def test_ea_incoming_particle_beam():
    """
    Test that gradients are tracking when incoming beam (being a `ParticleBeam`)
    requires grad.
    """
    ea = cheetah.Segment.from_ocelot(ares.cell, warnings=False).subcell(
        "AREASOLA1", "AREABSCR1"
    )
    incoming_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/astra/ACHIP_EA1_2021.1351.001"
    )

    incoming_beam.particles = nn.Parameter(incoming_beam.particles)

    outgoing_beam = ea.track(incoming_beam)

    assert outgoing_beam.particles.grad_fn is not None



"""
Tests that ensure that both beam types produce (roughly) the same results.
"""
import numpy as np
import torch

import cheetah


def test_from_twiss():
    """
    Test that a beams created from Twiss parameters have the same properties.
    """
    parameter_beam = cheetah.ParameterBeam.from_twiss(
        beta_x=torch.tensor(5.91253676811640894),
        alpha_x=torch.tensor(3.55631307633660354),
        emittance_x=torch.tensor(3.494768647122823e-09),
        beta_y=torch.tensor(5.91253676811640982),
        alpha_y=torch.tensor(2e-7),
        emittance_y=torch.tensor(3.497810737006068e-09),
        energy=torch.tensor(6e6),
    )
    particle_beam = cheetah.ParticleBeam.from_twiss(
        num_particles=torch.tensor(
            10_000_000
        ),  # Large number of particles reduces noise
        beta_x=torch.tensor(5.91253676811640894),
        alpha_x=torch.tensor(3.55631307633660354),
        emittance_x=torch.tensor(3.494768647122823e-09),
        beta_y=torch.tensor(5.91253676811640982),
        alpha_y=torch.tensor(2e-7),
        emittance_y=torch.tensor(3.497810737006068e-09),
        energy=torch.tensor(6e6),
    )

    assert np.isclose(parameter_beam.mu_x, particle_beam.mu_x, atol=1e-6)
    assert np.isclose(parameter_beam.mu_y, particle_beam.mu_y, atol=1e-6)
    assert np.isclose(parameter_beam.sigma_x, particle_beam.sigma_x, rtol=1e-3)
    assert np.isclose(parameter_beam.sigma_y, particle_beam.sigma_y, rtol=1e-3)
    assert np.isclose(parameter_beam.mu_xp, particle_beam.mu_xp, atol=1e-6)
    assert np.isclose(parameter_beam.mu_yp, particle_beam.mu_yp, atol=1e-6)
    assert np.isclose(parameter_beam.sigma_xp, particle_beam.sigma_xp, rtol=1e-3)
    assert np.isclose(parameter_beam.sigma_yp, particle_beam.sigma_yp, rtol=1e-3)
    assert np.isclose(parameter_beam.mu_s, particle_beam.mu_s)
    assert np.isclose(parameter_beam.sigma_s, particle_beam.sigma_s)
    assert np.isclose(parameter_beam.mu_p, particle_beam.mu_p)
    assert np.isclose(parameter_beam.sigma_p, particle_beam.sigma_p)


def test_drift():
    """Test that the drift output for both beam types is roughly the same."""

    # Set up lattice
    cheetah_drift = cheetah.Drift(length=torch.tensor(1.0))

    # Parameter beam
    incoming_parameter_beam = cheetah.ParameterBeam.from_astra(
        "benchmark/astra/ACHIP_EA1_2021.1351.001"
    )
    outgoing_parameter_beam = cheetah_drift.track(incoming_parameter_beam)

    # Particle beam
    incoming_particle_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/astra/ACHIP_EA1_2021.1351.001"
    )
    outgoing_particle_beam = cheetah_drift.track(incoming_particle_beam)

    # Compare
    assert np.isclose(outgoing_parameter_beam.energy, outgoing_particle_beam.energy)
    assert np.isclose(
        outgoing_parameter_beam.mu_x, outgoing_particle_beam.mu_x, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.mu_y, outgoing_particle_beam.mu_y, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.sigma_x, outgoing_particle_beam.sigma_x, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.sigma_y, outgoing_particle_beam.sigma_y, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.mu_xp, outgoing_particle_beam.mu_xp, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.mu_yp, outgoing_particle_beam.mu_yp, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.sigma_xp, outgoing_particle_beam.sigma_xp, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.sigma_yp, outgoing_particle_beam.sigma_yp, rtol=1e-2
    )


def test_quadrupole():
    """Test that the quadrupole output for both beam types is roughly the same."""

    # Set up lattice
    cheetah_quadrupole = cheetah.Quadrupole(
        length=torch.tensor(0.15), k1=torch.tensor(4.2)
    )

    # Parameter beam
    incoming_parameter_beam = cheetah.ParameterBeam.from_astra(
        "benchmark/astra/ACHIP_EA1_2021.1351.001"
    )
    outgoing_parameter_beam = cheetah_quadrupole.track(incoming_parameter_beam)

    # Particle beam
    incoming_particle_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/astra/ACHIP_EA1_2021.1351.001"
    )
    outgoing_particle_beam = cheetah_quadrupole.track(incoming_particle_beam)

    # Compare
    assert np.isclose(outgoing_parameter_beam.energy, outgoing_particle_beam.energy)
    assert np.isclose(
        outgoing_parameter_beam.mu_x, outgoing_particle_beam.mu_x, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.mu_y, outgoing_particle_beam.mu_y, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.sigma_x, outgoing_particle_beam.sigma_x, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.sigma_y, outgoing_particle_beam.sigma_y, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.mu_xp, outgoing_particle_beam.mu_xp, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.mu_yp, outgoing_particle_beam.mu_yp, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.sigma_xp, outgoing_particle_beam.sigma_xp, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.sigma_yp, outgoing_particle_beam.sigma_yp, rtol=1e-2
    )


def test_cavity_from_astra():
    """
    Test that the cavity output for both beam types is roughly the same. This test uses
    a beam converted from an ASTRA beam file.
    """

    # Set up lattice
    cheetah_cavity = cheetah.Cavity(
        length=torch.tensor(1.0377),
        voltage=torch.tensor(0.01815975e9),
        frequency=torch.tensor(1.3e9),
        phase=torch.tensor(0.0),
    )

    # Parameter beam
    incoming_parameter_beam = cheetah.ParameterBeam.from_astra(
        "benchmark/astra/ACHIP_EA1_2021.1351.001"
    )
    outgoing_parameter_beam = cheetah_cavity.track(incoming_parameter_beam)

    # Particle beam
    incoming_particle_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/astra/ACHIP_EA1_2021.1351.001"
    )
    outgoing_particle_beam = cheetah_cavity.track(incoming_particle_beam)

    # Compare
    assert np.isclose(
        outgoing_parameter_beam.beta_x, outgoing_particle_beam.beta_x, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.alpha_x, outgoing_particle_beam.alpha_x, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.beta_y, outgoing_particle_beam.beta_y, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.alpha_y, outgoing_particle_beam.alpha_y, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.emittance_x, outgoing_particle_beam.emittance_x
    )
    assert np.isclose(
        outgoing_parameter_beam.emittance_y, outgoing_particle_beam.emittance_y
    )
    assert np.isclose(outgoing_parameter_beam.energy, outgoing_particle_beam.energy)
    assert np.isclose(
        outgoing_parameter_beam.mu_x, outgoing_particle_beam.mu_x, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.mu_y, outgoing_particle_beam.mu_y, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.sigma_x, outgoing_particle_beam.sigma_x, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.sigma_y, outgoing_particle_beam.sigma_y, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.mu_xp, outgoing_particle_beam.mu_xp, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.mu_yp, outgoing_particle_beam.mu_yp, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.sigma_xp, outgoing_particle_beam.sigma_xp, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.sigma_yp, outgoing_particle_beam.sigma_yp, rtol=1e-2
    )


def test_cavity_from_twiss():
    """
    Test that the cavity output for both beam types is roughly the same. This test uses
    a beam generated from Twiss parameters.
    """

    # Set up lattice
    cheetah_cavity = cheetah.Cavity(
        length=torch.tensor(1.0377),
        voltage=torch.tensor(0.01815975e9),
        frequency=torch.tensor(1.3e9),
        phase=torch.tensor(0.0),
    )

    # Parameter beam
    incoming_parameter_beam = cheetah.ParameterBeam.from_twiss(
        beta_x=torch.tensor(5.91253677),
        alpha_x=torch.tensor(3.55631308),
        beta_y=torch.tensor(5.91253677),
        alpha_y=torch.tensor(3.55631308),
        emittance_x=torch.tensor(3.494768647122823e-09),
        emittance_y=torch.tensor(3.497810737006068e-09),
        energy=torch.tensor(6e6),
    )
    outgoing_parameter_beam = cheetah_cavity.track(incoming_parameter_beam)

    # Particle beam
    incoming_particle_beam = cheetah.ParticleBeam.from_twiss(
        beta_x=torch.tensor(5.91253677),
        alpha_x=torch.tensor(3.55631308),
        beta_y=torch.tensor(5.91253677),
        alpha_y=torch.tensor(3.55631308),
        emittance_x=torch.tensor(3.494768647122823e-09),
        emittance_y=torch.tensor(3.497810737006068e-09),
        energy=torch.tensor(6e6),
    )
    outgoing_particle_beam = cheetah_cavity.track(incoming_particle_beam)

    # Compare
    assert np.isclose(
        outgoing_parameter_beam.beta_x, outgoing_particle_beam.beta_x, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.alpha_x, outgoing_particle_beam.alpha_x, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.beta_y, outgoing_particle_beam.beta_y, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.alpha_y, outgoing_particle_beam.alpha_y, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.emittance_x, outgoing_particle_beam.emittance_x
    )
    assert np.isclose(
        outgoing_parameter_beam.emittance_y, outgoing_particle_beam.emittance_y
    )
    assert np.isclose(outgoing_parameter_beam.energy, outgoing_particle_beam.energy)
    assert np.isclose(
        outgoing_parameter_beam.mu_x, outgoing_particle_beam.mu_x, atol=1e-6
    )
    assert np.isclose(
        outgoing_parameter_beam.mu_y, outgoing_particle_beam.mu_y, atol=1e-6
    )
    assert np.isclose(
        outgoing_parameter_beam.sigma_x, outgoing_particle_beam.sigma_x, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.sigma_y, outgoing_particle_beam.sigma_y, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.mu_xp, outgoing_particle_beam.mu_xp, atol=1e-6
    )
    assert np.isclose(
        outgoing_parameter_beam.mu_yp, outgoing_particle_beam.mu_yp, atol=1e-6
    )
    assert np.isclose(
        outgoing_parameter_beam.sigma_xp, outgoing_particle_beam.sigma_xp, rtol=1e-2
    )
    assert np.isclose(
        outgoing_parameter_beam.sigma_yp, outgoing_particle_beam.sigma_yp, rtol=1e-2
    )



import torch

from cheetah import Drift, ParameterBeam, ParticleBeam


def test_diverging_parameter_beam():
    """
    Test that that a parameter beam with sigma_xp > 0 and sigma_yp > 0 increases in
    size in both dimensions when travelling through a drift section.
    """
    drift = Drift(length=torch.tensor(1.0))
    incoming_beam = ParameterBeam.from_parameters(
        sigma_xp=torch.tensor(2e-7), sigma_yp=torch.tensor(2e-7)
    )
    outgoing_beam = drift(incoming_beam)

    assert outgoing_beam.sigma_x > incoming_beam.sigma_x
    assert outgoing_beam.sigma_y > incoming_beam.sigma_y


def test_diverging_particle_beam():
    """
    Test that that a particle beam with sigma_xp > 0 and sigma_yp > 0 increases in
    size in both dimensions when travelling through a drift section.
    """
    drift = Drift(length=torch.tensor(1.0))
    incoming_beam = ParticleBeam.from_parameters(
        num_particles=torch.tensor(1000),
        sigma_xp=torch.tensor(2e-7),
        sigma_yp=torch.tensor(2e-7),
    )
    outgoing_beam = drift(incoming_beam)

    assert outgoing_beam.sigma_x > incoming_beam.sigma_x
    assert outgoing_beam.sigma_y > incoming_beam.sigma_y



import numpy as np
import pytest
import torch

import cheetah


@pytest.mark.xfail  # TODO: Fix this
def test_drift_end():
    """
    Test that at the end of a split drift the result is the same as at the end of the
    original drift.
    """
    original_drift = cheetah.Drift(length=torch.tensor(2.0))
    split_drift = cheetah.Segment(original_drift.split(resolution=torch.tensor(0.1)))

    incoming_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/astra/ACHIP_EA1_2021.1351.001"
    )

    outgoing_beam_original = original_drift.track(incoming_beam)
    outgoing_beam_split = split_drift.track(incoming_beam)

    assert np.allclose(outgoing_beam_original.particles, outgoing_beam_split.particles)


@pytest.mark.xfail  # TODO: Fix this
def test_quadrupole_end():
    """
    Test that at the end of a split quadrupole the result is the same as at the end of
    the original quadrupole.
    """
    original_quadrupole = cheetah.Quadrupole(
        length=torch.tensor(0.2), k1=torch.tensor(4.2)
    )
    split_quadrupole = cheetah.Segment(
        original_quadrupole.split(resolution=torch.tensor(0.01))
    )

    incoming_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/astra/ACHIP_EA1_2021.1351.001"
    )

    outgoing_beam_original = original_quadrupole.track(incoming_beam)
    outgoing_beam_split = split_quadrupole.track(incoming_beam)

    assert np.allclose(outgoing_beam_original.particles, outgoing_beam_split.particles)


def test_cavity_end():
    """
    Test that at the end of a split cavity the result is the same as at the end of
    the original cavity.
    """
    original_cavity = cheetah.Cavity(
        length=torch.tensor(1.0377),
        voltage=torch.tensor(0.01815975e9),
        frequency=torch.tensor(1.3e9),
        phase=torch.tensor(0.0),
    )
    split_cavity = cheetah.Segment(original_cavity.split(resolution=torch.tensor(0.1)))

    incoming_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/astra/ACHIP_EA1_2021.1351.001"
    )

    outgoing_beam_original = original_cavity.track(incoming_beam)
    outgoing_beam_split = split_cavity.track(incoming_beam)

    assert np.allclose(outgoing_beam_original.particles, outgoing_beam_split.particles)


def test_solenoid_end():
    """
    Test that at the end of a split solenoid the result is the same as at the end of
    the original solenoid.
    """
    original_solenoid = cheetah.Solenoid(length=torch.tensor(0.2), k=torch.tensor(4.2))
    split_solenoid = cheetah.Segment(
        original_solenoid.split(resolution=torch.tensor(0.01))
    )

    incoming_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/astra/ACHIP_EA1_2021.1351.001"
    )

    outgoing_beam_original = original_solenoid.track(incoming_beam)
    outgoing_beam_split = split_solenoid.track(incoming_beam)

    assert np.allclose(outgoing_beam_original.particles, outgoing_beam_split.particles)


def test_dipole_end():
    """
    Test that at the end of a split dipole the result is the same as at the end of
    the original dipole.
    """
    original_dipole = cheetah.Dipole(length=torch.tensor(0.2), angle=torch.tensor(4.2))
    split_dipole = cheetah.Segment(original_dipole.split(resolution=torch.tensor(0.01)))

    incoming_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/astra/ACHIP_EA1_2021.1351.001"
    )

    outgoing_beam_original = original_dipole.track(incoming_beam)
    outgoing_beam_split = split_dipole.track(incoming_beam)

    assert np.allclose(outgoing_beam_original.particles, outgoing_beam_split.particles)


def test_undulator_end():
    """
    Test that at the end of a split undulator the result is the same as at the end of
    the original undulator.
    """
    original_undulator = cheetah.Undulator(length=torch.tensor(3.142))
    split_undulator = cheetah.Segment(
        original_undulator.split(resolution=torch.tensor(0.1))
    )

    incoming_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/astra/ACHIP_EA1_2021.1351.001"
    )

    outgoing_beam_original = original_undulator.track(incoming_beam)
    outgoing_beam_split = split_undulator.track(incoming_beam)

    assert np.allclose(outgoing_beam_original.particles, outgoing_beam_split.particles)


@pytest.mark.xfail  # TODO: Fix this
def test_horizontal_corrector_end():
    """
    Test that at the end of a split horizontal corrector the result is the same as at
    the end of the original horizontal corrector.
    """
    original_horizontal_corrector = cheetah.HorizontalCorrector(
        length=torch.tensor(0.2), angle=torch.tensor(4.2)
    )
    split_horizontal_corrector = cheetah.Segment(
        original_horizontal_corrector.split(resolution=torch.tensor(0.01))
    )

    incoming_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/astra/ACHIP_EA1_2021.1351.001"
    )

    outgoing_beam_original = original_horizontal_corrector.track(incoming_beam)
    outgoing_beam_split = split_horizontal_corrector.track(incoming_beam)

    assert np.allclose(outgoing_beam_original.particles, outgoing_beam_split.particles)


@pytest.mark.xfail  # TODO: Fix this
def test_vertical_corrector_end():
    """
    Test that at the end of a split vertical corrector the result is the same as at
    the end of the original vertical corrector.
    """
    original_vertical_corrector = cheetah.VerticalCorrector(
        length=torch.tensor(0.2), angle=torch.tensor(4.2)
    )
    split_vertical_corrector = cheetah.Segment(
        original_vertical_corrector.split(resolution=torch.tensor(0.01))
    )

    incoming_beam = cheetah.ParticleBeam.from_astra(
        "benchmark/astra/ACHIP_EA1_2021.1351.001"
    )

    outgoing_beam_original = original_vertical_corrector.track(incoming_beam)
    outgoing_beam_split = split_vertical_corrector.track(incoming_beam)

    assert np.allclose(outgoing_beam_original.particles, outgoing_beam_split.particles)



import pytest

from cheetah import Screen, Segment

from .resources import ARESlatticeStage3v1_9 as ares


@pytest.mark.parametrize(
    "name",
    [
        "ARLIBSCX1",
        "ARLIBSCR1",
        "ARLIBSCR2",
        "ARLIBSCR3",
        "AREABSCR1",
        "ARMRBSCR1",
        "ARMRBSCR2",
        "ARMRBSCR3",
        "ARBCBSCE1",
        "ARDLBSCR1",
        "ARDLBSCE1",
        "ARSHBSCE2",
        "ARSHBSCE1",
    ],
)
def test_screen_conversion(name: str):
    """
    Test on the example of the ARES lattice that all screens are correctly converted to
    `cheetah.Screen`.
    ˚"""
    segment = Segment.from_ocelot(ares.cell)
    screen = getattr(segment, name)
    assert isinstance(screen, Screen)



import numpy as np

import cheetah


def test_astra_to_parameter_beam():
    """Test that Astra beams are correctly loaded into parameter beams."""
    beam = cheetah.ParameterBeam.from_astra("benchmark/astra/ACHIP_EA1_2021.1351.001")

    # True values taken from version of Cheetah that is belived to be correct (v0.5.19)
    assert np.allclose(beam.mu_x, 8.24126345833065e-07)
    assert np.allclose(beam.mu_xp, 5.988477624896404e-08)
    assert np.allclose(beam.mu_y, -1.7276204289373709e-06)
    assert np.allclose(beam.mu_yp, -1.1746412553748087e-07)
    assert np.allclose(beam.sigma_x, 0.00017489789752289653)
    assert np.allclose(beam.sigma_xp, 3.679402198031312e-06)
    assert np.allclose(beam.sigma_y, 0.00017519544053357095)
    assert np.allclose(beam.sigma_yp, 3.6941000871593133e-06)
    assert np.allclose(beam.sigma_s, 8.011552381503861e-06)
    assert np.allclose(beam.sigma_p, 0.0022804534528404474)
    assert np.allclose(beam.energy, 107315902.44394557)


def test_astra_to_particle_beam():
    """Test that Astra beams are correctly loaded into particle beams."""
    beam = cheetah.ParticleBeam.from_astra("benchmark/astra/ACHIP_EA1_2021.1351.001")

    # True values taken from version of Cheetah that is belived to be correct (v0.5.19)
    assert beam.num_particles == 100_000
    assert np.allclose(beam.mu_x, 8.24126345833065e-07)
    assert np.allclose(beam.mu_xp, 5.988477624896404e-08)
    assert np.allclose(beam.mu_y, -1.7276204289373709e-06)
    assert np.allclose(beam.mu_yp, -1.1746412553748087e-07)
    assert np.allclose(beam.sigma_x, 0.00017489789752289653)
    assert np.allclose(beam.sigma_xp, 3.679402198031312e-06)
    assert np.allclose(beam.sigma_y, 0.00017519544053357095)
    assert np.allclose(beam.sigma_yp, 3.6941000871593133e-06)
    assert np.allclose(beam.sigma_s, 8.011552381503861e-06)
    assert np.allclose(beam.sigma_p, 0.0022804534528404474)
    assert np.allclose(beam.energy, 107315902.44394557)



import torch

import cheetah


def test_bmad_tutorial():
    """Test importing the lattice example file from the Bmad and Tao tutorial."""
    file_path = "tests/resources/bmad_tutorial_lattice.bmad"
    converted = cheetah.Segment.from_bmad(file_path)
    converted.name = "bmad_tutorial"

    correct = cheetah.Segment(
        [
            cheetah.Drift(length=torch.tensor(0.5), name="d"),
            cheetah.Dipole(
                length=torch.tensor(0.5), e1=torch.tensor(0.1), name="b"
            ),  # TODO: What are g and dg?
            cheetah.Quadrupole(
                length=torch.tensor(0.6), k1=torch.tensor(0.23), name="q"
            ),
        ],
        name="bmad_tutorial",
    )

    assert converted.name == correct.name
    assert [element.name for element in converted.elements] == [
        element.name for element in correct.elements
    ]
    assert converted.d.length == correct.d.length
    assert converted.b.length == correct.b.length
    assert converted.b.e1 == correct.b.e1
    assert converted.q.length == correct.q.length
    assert converted.q.k1 == correct.q.k1



"""Utility functions for creating transfer maps for the elements."""

from typing import Optional, Union

import torch
from scipy import constants

REST_ENERGY = torch.tensor(
    constants.electron_mass
    * constants.speed_of_light**2
    / constants.elementary_charge
)  # electron mass


def rotation_matrix(
    angle: torch.Tensor, device: Union[str, torch.device] = "auto"
) -> torch.Tensor:
    """Rotate the transfer map in x-y plane

    :param angle: Rotation angle in rad, for example `angle = np.pi/2` for vertical =
        dipole.
    :param device: Device used for tracking, by default "auto".
    :return: Rotation matrix to be multiplied to the element's transfer matrix.
    """
    if device == "auto":
        device = "cuda" if torch.cuda.is_available() else "cpu"
    cs = torch.cos(angle)
    sn = torch.sin(angle)

    tm = torch.eye(7, dtype=torch.float32, device=device)
    tm[0, 0] = cs
    tm[0, 2] = sn
    tm[1, 1] = cs
    tm[1, 3] = sn
    tm[2, 0] = -sn
    tm[2, 2] = cs
    tm[3, 1] = -sn
    tm[3, 3] = cs

    return tm


def base_rmatrix(
    length: torch.Tensor,
    k1: torch.Tensor,
    hx: torch.Tensor,
    tilt: Optional[torch.Tensor] = None,
    energy: Optional[torch.Tensor] = None,
    device: Union[str, torch.device] = "auto",
) -> torch.Tensor:
    """
    Create a universal transfer matrix for a beamline element.

    :param length: Length of the element in m.
    :param k1: Quadrupole strength in 1/m**2.
    :param hx: Curvature (1/radius) of the element in 1/m**2.
    :param tilt: Roation of the element relative to the longitudinal axis in rad.
    :param energy: Beam energy in eV.
    :param device: Device where the transfer matrix is created. If "auto", the device
        is selected automatically.
    :return: Transfer matrix for the element.
    """

    tilt = tilt if tilt is not None else torch.tensor(0.0)
    energy = energy if energy is not None else torch.tensor(0.0)

    if device == "auto":
        device = "cuda" if torch.cuda.is_available() else "cpu"
    gamma = energy / REST_ENERGY
    igamma2 = 1 / gamma**2 if gamma != 0 else torch.tensor(0.0)

    beta = torch.sqrt(1 - igamma2)

    if k1 == 0:
        k1 = k1 + torch.tensor(1e-10, device=device)  # Avoid division by zero
    kx2 = k1 + hx**2
    ky2 = -k1
    kx = torch.sqrt(torch.complex(kx2, torch.tensor(0.0)))
    ky = torch.sqrt(torch.complex(ky2, torch.tensor(0.0)))
    cx = torch.cos(kx * length).real
    cy = torch.cos(ky * length).real
    sy = (torch.sin(ky * length) / ky).real if ky != 0 else length

    sx = (torch.sin(kx * length) / kx).real
    dx = hx / kx2 * (1.0 - cx)
    r56 = hx**2 * (length - sx) / kx2 / beta**2

    r56 -= length / beta**2 * igamma2

    R = torch.eye(7, dtype=torch.float32, device=device)
    R[0, 0] = cx
    R[0, 1] = sx
    R[0, 5] = dx / beta
    R[1, 0] = -kx2 * sx
    R[1, 1] = cx
    R[1, 5] = sx * hx / beta
    R[2, 2] = cy
    R[2, 3] = sy
    R[3, 2] = -ky2 * sy
    R[3, 3] = cy
    R[4, 0] = sx * hx / beta
    R[4, 1] = dx / beta
    R[4, 5] = r56

    # Rotate the R matrix for skew / vertical magnets
    if tilt != 0:
        R = torch.matmul(torch.matmul(rotation_matrix(-tilt), R), rotation_matrix(tilt))
    return R


def misalignment_matrix(
    misalignment: torch.Tensor, device: torch.device
) -> tuple[torch.Tensor, torch.Tensor]:
    """Shift the beam for tracking beam through misaligned elements"""
    R_exit = torch.eye(7, dtype=torch.float32, device=device)
    R_exit[0, 6] = misalignment[0]
    R_exit[2, 6] = misalignment[1]

    R_entry = torch.eye(7, dtype=torch.float32, device=device)
    R_entry[0, 6] = -misalignment[0]
    R_entry[2, 6] = -misalignment[1]

    return R_exit, R_entry  # TODO: This order is confusing, should be entry, exit



import numpy as np
import torch

from cheetah import ParameterBeam


def test_create_from_parameters():
    """
    Test that a `ParameterBeam` created from parameters actually has those parameters.
    """
    beam = ParameterBeam.from_parameters(
        mu_x=torch.tensor(1e-5),
        mu_xp=torch.tensor(1e-7),
        mu_y=torch.tensor(2e-5),
        mu_yp=torch.tensor(2e-7),
        sigma_x=torch.tensor(1.75e-7),
        sigma_xp=torch.tensor(2e-7),
        sigma_y=torch.tensor(1.75e-7),
        sigma_yp=torch.tensor(2e-7),
        sigma_s=torch.tensor(0.000001),
        sigma_p=torch.tensor(0.000001),
        cor_x=torch.tensor(0.0),
        cor_y=torch.tensor(0.0),
        cor_s=torch.tensor(0.0),
        energy=torch.tensor(1e7),
    )

    assert np.isclose(beam.mu_x, 1e-5)
    assert np.isclose(beam.mu_xp, 1e-7)
    assert np.isclose(beam.mu_y, 2e-5)
    assert np.isclose(beam.mu_yp, 2e-7)
    assert np.isclose(beam.sigma_x, 1.75e-7)
    assert np.isclose(beam.sigma_xp, 2e-7)
    assert np.isclose(beam.sigma_y, 1.75e-7)
    assert np.isclose(beam.sigma_yp, 2e-7)
    assert np.isclose(beam.sigma_s, 0.000001)
    assert np.isclose(beam.sigma_p, 0.000001)
    assert np.isclose(beam.energy, 1e7)


def test_transform_to():
    """
    Test that a `ParameterBeam` transformed to new parameters actually has those new
    parameters.
    """
    original_beam = ParameterBeam.from_parameters()
    transformed_beam = original_beam.transformed_to(
        mu_x=torch.tensor(1e-5),
        mu_xp=torch.tensor(1e-7),
        mu_y=torch.tensor(2e-5),
        mu_yp=torch.tensor(2e-7),
        sigma_x=torch.tensor(1.75e-7),
        sigma_xp=torch.tensor(2e-7),
        sigma_y=torch.tensor(1.75e-7),
        sigma_yp=torch.tensor(2e-7),
        sigma_s=torch.tensor(0.000001),
        sigma_p=torch.tensor(0.000001),
        energy=torch.tensor(1e7),
    )

    assert isinstance(transformed_beam, ParameterBeam)
    assert np.isclose(transformed_beam.mu_x, 1e-5)
    assert np.isclose(transformed_beam.mu_xp, 1e-7)
    assert np.isclose(transformed_beam.mu_y, 2e-5)
    assert np.isclose(transformed_beam.mu_yp, 2e-7)
    assert np.isclose(transformed_beam.sigma_x, 1.75e-7)
    assert np.isclose(transformed_beam.sigma_xp, 2e-7)
    assert np.isclose(transformed_beam.sigma_y, 1.75e-7)
    assert np.isclose(transformed_beam.sigma_yp, 2e-7)
    assert np.isclose(transformed_beam.sigma_s, 0.000001)
    assert np.isclose(transformed_beam.sigma_p, 0.000001)
    assert np.isclose(transformed_beam.energy, 1e7)


def test_from_twiss_to_twiss():
    """
    Test that a `ParameterBeam` created from twiss parameters actually has those
    parameters.
    """
    beam = ParameterBeam.from_twiss(
        beta_x=torch.tensor(5.91253676811640894),
        alpha_x=torch.tensor(3.55631307633660354),
        emittance_x=torch.tensor(3.494768647122823e-09),
        beta_y=torch.tensor(5.91253676811640982),
        alpha_y=torch.tensor(2e-7),
        emittance_y=torch.tensor(3.497810737006068e-09),
        energy=torch.tensor(6e6),
    )

    assert np.isclose(beam.beta_x, 5.91253676811640894)
    assert np.isclose(beam.alpha_x, 3.55631307633660354)
    assert np.isclose(beam.emittance_x, 3.494768647122823e-09)
    assert np.isclose(beam.beta_y, 5.91253676811640982)
    assert np.isclose(beam.alpha_y, 2e-7)
    assert np.isclose(beam.emittance_y, 3.497810737006068e-09)
    assert np.isclose(beam.energy, 6e6)



class UniqueNameGenerator:
    """Generates a unique name given a prefix."""

    def __init__(self, prefix):
        self._prefix = prefix
        self._counter = 0

    def __call__(self):
        name = f"{self._prefix}_{self._counter}"
        self._counter += 1
        return name
